"""
Contains the class `ModelStructure`.
"""


from .voltage import load_voltage_model, VoltageFunction
from .basisfunctions import extract_basis_functions
from .dataclasses import BasisFunction
from .typeddicts import VoltageSocData


class ModelStructure:
    """This class allows battery model identification and simulation employing
    the proposed model structure using the input-output (IO) representation in
    the linear parameter-varying (LPV) framework.

    Attributes
    ----------
    battery_capacity : float
        Capacity of the battery being modelled.
    sampling_period: float
        Model sampling time.
    """

    battery_capacity: float
    sampling_period: float

    emf_function: VoltageFunction
    hysteresis_function: VoltageFunction | None

    basis_functions: list[BasisFunction]
    hysteresis_basis_functions: list[BasisFunction]


    def __init__(self, battery_capacity: float, sampling_period: float):
        self.battery_capacity = battery_capacity
        self.sampling_period = sampling_period

        self.hysteresis_function = None

        self.basis_functions = []
        self.hysteresis_basis_functions = []


    def add_emf_function(self, voltage_soc_data: VoltageSocData):
        """Add EMF function used to decompose battery voltage into
        overpotentials and vice versa."""
        #
        self.emf_function = load_voltage_model(voltage_soc_data)


    def add_hysteresis_function(self, voltage_soc_data: VoltageSocData):
        """Add hysteresis function to be used as second model input."""
        #
        self.hysteresis_function = load_voltage_model(voltage_soc_data)


    def add_basis_functions(self, basis_function_strings: list[str],
                            hysteresis_basis_function_strings: list[str] | None = None):
        """Add basis functions by extracting signal variable, the desired
        operation and its arguments."""
        self.basis_functions = extract_basis_functions(basis_function_strings)
        if hysteresis_basis_function_strings is not None:
            hbfs = hysteresis_basis_function_strings
            self.hysteresis_basis_functions = extract_basis_functions(hbfs)



"""
This file contains the main function(s) to construct
a voltage (e.g., EMF) model.
"""


import numpy as np
from scipy.interpolate import PchipInterpolator

from .dataclasses import VoltageFunction
from .typeddicts import VoltageSocData


def create_soc_interpolator(soc_values: np.ndarray, y_values: np.ndarray,
                            extrapolate=False) -> PchipInterpolator:
    """Create interpolation function from y--soc data. Note that
    y can be either voltage or temperature."""

    # If the SOC direction is descending, we sort it to
    # ascending first.
    if any(np.diff(soc_values) < 0):
        # We reverse the lists
        soc_values = soc_values[::-1]
        y_values = y_values[::-1]

    interpolator = PchipInterpolator(soc_values, y_values, extrapolate=extrapolate)
    return interpolator


def load_voltage_model(dataset: VoltageSocData):
    """Load temperature-dependent voltage model."""
    #
    dVdT_soc_function = None # pylint: disable=C0103
    reference_temperature = None
    if 'dVdT_values' in dataset:
        if 'reference_temperature_value' not in dataset:
            raise ValueError('Reference temperature value not specified.')
        # pylint: disable=C0103
        dVdT_soc_function = create_soc_interpolator(dataset['soc_values'], dataset['dVdT_values'])
        reference_temperature = dataset['reference_temperature_value']

    # We create and store interpolation functions for
    # the given Voltage--SOC data.
    voltage_soc_function = create_soc_interpolator(dataset['soc_values'], dataset['voltage_values'])

    return VoltageFunction(voltage_soc_function, dVdT_soc_function, reference_temperature)



"""Utilities related to model identification."""


from typing import Literal

import numpy as np

from .modelstructure import ModelStructure
from .dataclasses import Model
from .typeddicts import CurrentVoltageData
from .regression import setup_regression, combine_regression_problems, run_optimizer
from .basisfunctions import combine_symbols, generate_signals, generate_basis_function_signals, \
    generate_signal_trajectories


# pylint: disable=R0914
def setup_regression_problems(datasets: list[CurrentVoltageData],
                              model_structure: ModelStructure,
                              model_order: int, nonlinearity_order: int):
    """Generate regression matrices for the given dataset(s)."""
    # Shorthand for convenience
    ms = model_structure
    #
    regression_problems = []
    for dataset in datasets:
        signals = generate_signals(dataset, ms.battery_capacity, ms.sampling_period,
                                   ms.emf_function, ms.hysteresis_function)
        #
        bf_signal_vector = generate_basis_function_signals(ms.basis_functions, signals)
        hysteresis_bf_signal_vector = generate_basis_function_signals(ms.hysteresis_basis_functions,
                                                              signals)
        #
        io_signals = [ signals.find('v'), signals.find('i') ]
        if ms.hysteresis_function is not None:
            io_signals.append( signals.find('h') )
        #
        signal_tuple = io_signals, bf_signal_vector.signals, hysteresis_bf_signal_vector.signals
        signal_trajectories = generate_signal_trajectories(signal_tuple,
                                                            model_order, model_order)
        io_trajectories, p_trajectories, h_trajectories = signal_trajectories
        #
        basis_function_keys = combine_symbols(bf_signal_vector.get_symbols(), nonlinearity_order)
        hysteresis_basis_function_keys = hysteresis_bf_signal_vector.get_symbols()
        #
        regression_matrix, regressor_labels = setup_regression(io_trajectories,
                                                                p_trajectories,
                                                                h_trajectories,
                                                                basis_function_keys,
                                                                hysteresis_basis_function_keys)
        output_vector = np.array(signals.find('v').trajectory[model_order:])
        #
        regression_problems.append( (regression_matrix, output_vector) )
    #
    return regression_problems, regressor_labels


# pylint: disable=too-many-arguments, R0917
def identify_model(datasets: list[CurrentVoltageData] | CurrentVoltageData,
                   model_structure: ModelStructure,
                   model_order: int, nonlinearity_order: int,
                   optimizers: Literal['lasso.cvxopt', 'lasso.sklearn', 'ridge.sklearn'],
                   combining_strategy: Literal['concatenate', 'interleave']='interleave'):
    """Identify a battery model using the provided identification
    dataset along with the desired model order and basis-function
    complexity."""
    # Shorthand for convenience
    ms = model_structure
    #
    if isinstance(datasets, dict):
        datasets = [ datasets ]
    regression_problems, regressor_labels = setup_regression_problems(datasets,
                                                                      model_structure,
                                                                      model_order,
                                                                      nonlinearity_order)
    # Combine multiple regression problems
    regression_matrix, output_vector = combine_regression_problems(regression_problems,
                                                                    strategy=combining_strategy)
    #
    if not optimizers:
        raise ValueError('Unspecified optimization routine(s).')
    #
    model_estimate = np.array([])
    for optimizer in optimizers:
        #
        model_estimate = run_optimizer(regression_matrix, output_vector, optimizer.lower())
        #
        if 'lasso' in optimizer.lower():
            selected_regressors = np.abs(model_estimate) > 1e-5
            regression_matrix = regression_matrix[:, selected_regressors]  # type: ignore
            regressor_labels = np.array(regressor_labels)[selected_regressors]
            model_estimate = np.array(model_estimate)[selected_regressors]

    return Model(ms.battery_capacity,
                 ms.sampling_period,
                 model_order,
                 nonlinearity_order,
                 regressor_labels,
                 model_estimate,
                 ms.basis_functions,
                 ms.hysteresis_basis_functions,
                 ms.emf_function,
                 ms.hysteresis_function)



"""
Testing related to combination of basis functions from
user-provided strings.
"""

from pybatteryid.basisfunctions import combine_symbols

def test_s_combination():
    """Test combinations involving s variables."""
    symbols = ['s', '1/s']
    combinations = combine_symbols(symbols, 3)

    assert combinations == [('s',), ('1/s',), ('s', 's'), ('1/s', '1/s'),
                            ('s', 's', 's'), ('1/s', '1/s', '1/s')]

def test_d_combination():
    """Test combinations involving d variables."""
    symbols = ['d[0.01,0.99]', 'd[0.001,0.999]']
    combinations = combine_symbols(symbols, 10)

    assert combinations == [('d[0.01,0.99]',), ('d[0.001,0.999]',)]

def test_s_and_d_combination():
    """Test combinations involving s and d variables."""
    symbols = ['s', '1/s', 'd[0.001,0.999]']
    combinations = combine_symbols(symbols, 2)

    assert combinations == [('s',), ('1/s',), ('d[0.001,0.999]',), ('s', 's'),
                            ('s', 'd[0.001,0.999]'), ('1/s', '1/s'), ('1/s', 'd[0.001,0.999]')]



"""
Testing related to basis functions operation from
user-provided strings.
"""

import math as m

from pybatteryid.basisfunctions import extract_basis_functions, generate_basis_function_signals
from pybatteryid.dataclasses import SignalVector


def test_no_operation(scheduling_signals: SignalVector):
    """No operation."""

    s = scheduling_signals.find('s')
    #
    strings = ['s']
    basis_functions = extract_basis_functions(strings)
    signal_vector = generate_basis_function_signals(basis_functions, scheduling_signals)
    #
    result = signal_vector.find(strings[0])

    assert result.symbol == s.symbol
    assert result.trajectory == s.trajectory

def test_inverse_operation(scheduling_signals: SignalVector):
    """Inverse operation."""

    s = scheduling_signals.find('s')
    #
    strings = ['1/s']
    basis_functions = extract_basis_functions(strings)
    signal_vector = generate_basis_function_signals(basis_functions, scheduling_signals)
    #
    result = signal_vector.find(strings[0])

    assert result.symbol == strings[0]
    assert result.trajectory == [1 / x for x in s.trajectory]

def test_log_operation(scheduling_signals: SignalVector):
    """Logarithmic operation."""

    s = scheduling_signals.find('s')
    #
    strings = ['log[s]']
    basis_functions = extract_basis_functions(strings)
    signal_vector = generate_basis_function_signals(basis_functions, scheduling_signals)
    #
    result = signal_vector.find(strings[0])

    assert result.symbol == strings[0]
    assert result.trajectory == [m.log(x) for x in s.trajectory]

def test_exp_sqrt_abs_operation(scheduling_signals: SignalVector):
    """Exponential square-root operation."""

    i = scheduling_signals.find('i')
    s = scheduling_signals.find('s')
    #
    strings = ['exp[0.05*sqrt[|i|]]', 'exp[-2.759*sqrt[0.5*s+140]]']
    basis_functions = extract_basis_functions(strings)
    signal_vector = generate_basis_function_signals(basis_functions, scheduling_signals)
    #
    result = signal_vector.find(strings[0])

    assert result.symbol == strings[0]
    assert result.trajectory == [m.exp(0.05*m.sqrt(abs(x))) for x in i.trajectory]

    result = signal_vector.find(strings[1])

    assert result.symbol == strings[1]
    assert result.trajectory == [m.exp(-2.759*m.sqrt(0.5*x+140)) for x in s.trajectory]

def test_exp_pow_abs_operation(scheduling_signals: SignalVector):
    """Exponential arbitrary power operation."""

    T = scheduling_signals.find('T') # pylint: disable=C0103
    i = scheduling_signals.find('i')
    #
    strings = ['exp[[0.00366*T+1]^-1]', 'exp[-2.759*[|-1.5*i+1890|]^+2.89]']
    basis_functions = extract_basis_functions(strings)
    signal_vector = generate_basis_function_signals(basis_functions, scheduling_signals)
    #
    result = signal_vector.find(strings[0])

    assert result.symbol == strings[0]
    assert result.trajectory == [m.exp((0.00366*x+1)**-1) for x in T.trajectory]

    result = signal_vector.find(strings[1])

    assert result.symbol == strings[1]
    assert result.trajectory == [m.exp(-2.759*(-1.5*x+1890)**2.89) for x in i.trajectory]



"""
A collection of classes for various typed dictionaries.
"""


from typing import TypedDict, NotRequired
from numpy.typing import NDArray


class VoltageSocData(TypedDict):
    """A dictionary for voltage--soc data."""
    soc_values: NDArray
    voltage_values: NDArray
    dVdT_values: NotRequired[NDArray]
    reference_temperature_value: NotRequired[float]


class CurrentVoltageData(TypedDict):
    """A dictionary for current--voltage data."""
    initial_soc: float
    time_values: NDArray
    current_values: NDArray
    voltage_values: NDArray
    temperature_values: NotRequired[NDArray]



"""Utilities regarding regression."""


from typing import Literal

import numpy as np

from cvxopt import matrix
from sklearn.linear_model import LassoCV, RidgeCV

from rich.panel import Panel
from rich.jupyter import print as rich_print

from .l1regls import l1regls


def setup_regression(io_trajectories: dict, p_trajectories: dict, h_trajectories: dict,
                     basis_function_keys: list[list[str]],
                     hysteresis_basis_function_keys: list[str]):
    """Setup regression problem by constructing a regression matrix."""
    phi = []
    column_labels = []
    #
    for key, trajectory in io_trajectories.items():
        # extract delay from key
        symbol = key.split("(")[0]
        delay = "(" + key.split("(")[-1]
        #
        phi.append( trajectory )
        column_labels.append(key)

        if symbol == 'h':
            for basis_key in hysteresis_basis_function_keys:
                #
                h_trajectory = h_trajectories[basis_key + delay]
                phi.append( np.prod([trajectory, h_trajectory], axis=0) )

                combined_key = key + "×" + basis_key + delay
                column_labels.append( combined_key )
        else:
            for basis_key in basis_function_keys:
                #
                p_trajectories_list = [ p_trajectories[term + delay] for term in basis_key ]
                phi.append( np.prod([trajectory, *p_trajectories_list], axis=0) )

                combined_key = key + "×" + "×".join([ term + delay for term in basis_key ])
                column_labels.append( combined_key )
    #
    phi = np.array(phi).T
    #
    return phi, np.array(column_labels)


# pylint: disable=too-many-locals
def combine_regression_problems(problems: list[tuple[np.ndarray, ...]],
                                strategy: Literal['concatenate', 'interleave']='interleave'):
    """Combine regression problems A1*x=y1, A2*x=y2, and so on."""
    # possible schemes: append, interleave
    if strategy == 'interleave':
        # Lengths for all regression problems
        problem_sizes = sorted(set(len(problem[1]) for problem in problems))
        # Since the regression problems may not of same size, we
        # split the problems into subproblems such that in each split,
        # the problems are of the same length.
        regression_matrix_parts = []
        output_vector_parts = []
        # We split the problems into subproblems with
        # equal lengths
        for i, problem_size in enumerate(problem_sizes):
            start_idx = problem_sizes[i - 1] if i > 0 else 0
            end_idx = problem_size
            #
            problems_split = [(problem[0][start_idx:end_idx], problem[1][start_idx:end_idx])
                              for problem in problems if len(problem[1][start_idx:end_idx])]
            # Number of problems in the split (referred to as p' in the paper)
            number_of_problems = len(problems_split)
            # Remove all unwanted rows and elements from the regression matrices
            # and the output vectors, respectively.
            regression_matrices = [problem[0][i::number_of_problems]
                                   for i, problem in enumerate(problems_split)]
            output_vectors = [problem[1][i::number_of_problems]
                              for i, problem in enumerate(problems_split)]

            # Initialize final regression matrix and output vector.
            regression_matrix_split = np.empty_like(problems_split[0][0])
            output_vector_split = np.empty_like(problems_split[0][1])
            # Insert rows and elements alternatively (round-robin style) in the regression
            # matrix and output vector, respectively.
            for i, (regression_matrix, output_vector) in enumerate(zip(regression_matrices,
                                                                       output_vectors)):
                if len(regression_matrix) == 0:
                    continue
                regression_matrix_split[i::number_of_problems] = regression_matrix
                output_vector_split[i::number_of_problems] = output_vector
            #
            regression_matrix_parts.append(regression_matrix_split)
            output_vector_parts.append(output_vector_split)
            #
        final_regression_matrix = np.concatenate(regression_matrix_parts)
        final_output_vector = np.concatenate(output_vector_parts)
    elif strategy == 'concatenate':
        regression_matrices = [problem[0] for problem in problems]
        output_vectors = [problem[1] for problem in problems]
        #
        final_regression_matrix = np.concatenate(regression_matrices)
        final_output_vector = np.concatenate(output_vectors)
    else:
        raise ValueError('Invalid strategy to combine regression problems.')
    #
    rich_print(Panel(("Inverse condition number of regression matrix: "
                      f"{1 / np.linalg.cond(final_regression_matrix)}"
                      "\nDimensions of regression matrix: "
                      f"({final_regression_matrix.shape[0]} rows,"
                      f" {final_regression_matrix.shape[1]} columns)")))
    #
    return final_regression_matrix, final_output_vector


def run_optimizer(regression_matrix: np.ndarray, output_vector: np.ndarray, optimizer: str):
    """Run an optimization routine for a regression problem to
    estimate parameters."""
    #
    if optimizer == 'lasso.cvxopt':
        #
        rich_print(Panel(("Performing LASSO using `l1regls.py` from cvxopt.org "
                          "using lambda_1 = 1. See the following link for details:"
                          "\nhttps://cvxopt.org/examples/mlbook/l1regls.html")))
        #
        estimate = l1regls(matrix(regression_matrix), matrix(output_vector))
        estimate = np.array(estimate).flatten()
    elif optimizer == 'lassocv.sklearn':
        #
        rich_print(Panel(("Performing cross-validated LASSO using `sklearn` package.")))
        #
        alphas = np.linspace(1, 4, num=4 ) / (2 * regression_matrix.shape[0])
        model_fit = LassoCV(fit_intercept=False, alphas=alphas, max_iter=1000000000,
                            tol=0.01, cv=10, verbose=True).fit(regression_matrix, output_vector)
        estimate = np.array(model_fit.coef_).flatten()
        #
        rich_print("Solution found using [bold]lambda_1 = 2 * alpha * n_samples = "
                   f"{2 * model_fit.alpha_ * regression_matrix.shape[0]:.2f}")
    elif optimizer == 'ridgecv.sklearn':
        #
        rich_print(Panel("Performing cross-validated Ridge regression using `sklearn` package."))
        #
        alphas = np.logspace(-3, 0, num=4 )
        model_fit = RidgeCV(fit_intercept=False, cv=10, alphas=alphas).fit(regression_matrix,
                                                                           output_vector)
        estimate = model_fit.coef_
        #
        rich_print(f"Solution found using [bold]lambda_2 = alpha = {model_fit.alpha_}")
    else:
        raise ValueError('Unknown optimization routine.')
    #
    return estimate



"""
Several utilities for user's convenience.
"""


from dataclasses import asdict

import numpy as np

from scipy.integrate import trapezoid
from scipy.optimize import minimize_scalar
from scipy.spatial import ConvexHull # pylint: disable=E0611
from rich.jupyter import print as rich_print
from rich.table import Table

from .typeddicts import CurrentVoltageData
from .dataclasses import VoltageFunction, Model, BasisFunction
from .basisfunctions import generate_signals
from .plotter import plot_custom


# pylint: disable=too-many-locals
def analyze_temperature_soc_space(datasets: list[CurrentVoltageData], battery_capacity: float,
                                  sampling_period: float, emf_function: VoltageFunction,
                                  hysteresis_function: VoltageFunction | None = None):
    """Find the SOC range of given dataset(s) """
    #
    soc_temperature_tuples = []
    soc_values = []
    temperature_values = []
    for dataset in datasets:
        signals = generate_signals(dataset, battery_capacity, sampling_period,
                               emf_function, hysteresis_function)
        soc_signal = signals.find('s').trajectory
        temperature_signal = signals.find('T').trajectory
        #
        soc_temperature_tuples.append((soc_signal, temperature_signal))
        soc_values = np.concatenate((soc_values, soc_signal))
        temperature_values = np.concatenate((temperature_values, temperature_signal))
    #
    points = np.column_stack((soc_values, temperature_values))
    hull = ConvexHull(points)
    #
    simplex_tuples = []
    for simplex in hull.simplices:
        simplex_tuples.append((points[simplex, 0], points[simplex, 1]))
    #
    plot_custom(simplex_tuples + soc_temperature_tuples, xlabel=r'SOC [-]',
                ylabel=r'Temperature [$^\circ$C]', colors=['k'] * len(simplex_tuples),
                figsize=(6, 2), xaxis_reverse=True)


def invert_voltage_function(voltage_function: VoltageFunction, voltage: float,
                            temperature: float | None = None):
    """Determine SOC value corresponding to voltage value."""
    def _function(soc: float):
        return abs(voltage_function(soc, temperature) - voltage)
    #
    return minimize_scalar(_function, bounds=(0, 1)).x


# pylint: disable=too-many-locals, too-many-arguments, R0917
def analyze_dataset(dataset: CurrentVoltageData, battery_capacity: float,
                    sampling_period: float, emf_function: VoltageFunction,
                    hysteresis_function: VoltageFunction | None = None,
                    signals_to_plot: list[str] | None = None):
    """Print basic details of the given dataset such as SOC range,
    cycled capacity, voltage range, etc."""
    #
    signals = generate_signals(dataset, battery_capacity, sampling_period,
                               emf_function, hysteresis_function)
    # Print a table
    table = Table(title="Dataset analysis", show_header=False)
    # Number of samples
    number_of_samples = len(dataset['time_values'])
    table.add_row("Number of samples", f"{number_of_samples}")
    # Time in hours
    experiment_time = dataset['time_values'][-1] - dataset['time_values'][0]
    table.add_row("Experiment time [s], [h]",
                  f"{experiment_time:.3f}, {experiment_time / 3600:.3f}")
    # Cycled capacity
    cycled_capacity = trapezoid(dataset['current_values'])
    table.add_row("Extracted charge [C], [Ah]",
                  f"{-cycled_capacity:.3f}, {-cycled_capacity / 3600:.5f}")
    # Initial voltage value
    initial_voltage = dataset['voltage_values'][0]
    table.add_row("Initial voltage [V]", f"{initial_voltage:.5f}")
    # Voltage range
    voltage = dataset['voltage_values']
    table.add_row("Voltage range [V]", f"({min(voltage):.5f}, {max(voltage):.5f})")
    # SOC range
    soc = signals.find('s').trajectory
    table.add_row("SOC range [%]", f"({min(soc) * 100:.2f}, {max(soc) * 100:.2f})")
    # Temperature range
    if 'temperature_values' in dataset:
        temperature = dataset['temperature_values']
        table.add_row("Temperature range [°C]", f"({min(temperature):.2f}, {max(temperature):.2f})")
    #
    rich_print(table, markup=False)
    #
    if signals_to_plot:
        for symbol in signals_to_plot:
            signal_trajectory = signals.find(symbol).trajectory
            plot_custom([(dataset['time_values'], signal_trajectory)], figsize=(6, 2),
                        xlabel='Time', ylabel=symbol, colors=['darkblue'])


def save_model_to_file(model: Model, path_to_directory: str, description: str):
    """Save a model to file."""
    file_name = f'{description}_n,l={model.model_order},{model.nonlinearity_order}'
    file_path = f'{path_to_directory}/{file_name}'

    np.save(file_path, asdict(model)) # type: ignore


def load_model_from_file(file_path: str):
    """Load a model from file."""
    model_as_dict = np.load(f'{file_path}', allow_pickle=True).item()
    #
    try:
        model = Model(**model_as_dict)
        # Load basis functions
        model.basis_functions = [BasisFunction(**bf) for bf in model_as_dict['basis_functions']]
        model.hysteresis_basis_functions = [BasisFunction(**hbf)
                                            for hbf in model_as_dict['hysteresis_basis_functions']]
        # Load voltage functions
        model.emf_function = VoltageFunction(**model_as_dict['emf_function'])
        if model_as_dict['hysteresis_function'] is not None:
            model.hysteresis_function = VoltageFunction(**model_as_dict['hysteresis_function'])
    except Exception as exc:
        raise ValueError('Cannot load model.') from exc
    #
    return model


def print_model_details(model: Model):
    """Print model details."""
    # We print the details as a table.
    table = Table(title=(f"Model order = {model.model_order}; "
                         f"Nonlinearity order = {model.nonlinearity_order}; "
                         f"Number of model terms = {len(model.model_estimate)}"))
    #
    table.add_column("Model Term")
    table.add_column("Estimated Parameter")
    #
    for estimate, term in zip(model.model_estimate, model.model_terms):
        table.add_row(str(term), str(estimate))
    #
    rich_print(table, markup=False)



"""Voltage simulation utilities."""


import numpy as np

from .dataclasses import Model, Signal
from .typeddicts import CurrentVoltageData
from .basisfunctions import generate_signals, generate_basis_function_signals, \
    generate_signal_trajectories


# pylint: disable=too-many-locals
def simulate_model(model: Model, dataset: CurrentVoltageData):
    """Simulate the battery voltage using the provided model
    with the battery current as input."""
    #
    no_of_initial_values = len(dataset['voltage_values'])
    # Check if minimum number of initial values are provided.
    if no_of_initial_values < model.model_order:
        raise ValueError(f'At least {model.model_order} initial voltage value(s)'
                            ' should be provided.')
    #
    signal_vector = generate_signals(dataset, model.battery_capacity, model.sampling_period,
                                     model.emf_function, model.hysteresis_function)
    #
    bf_signal_vector = generate_basis_function_signals(model.basis_functions, signal_vector)
    hysteresis_bf_signal_vector = generate_basis_function_signals(model.hysteresis_basis_functions,
                                                                  signal_vector)
    #
    io_signals = [ signal_vector.find('v'), signal_vector.find('i') ]
    if model.hysteresis_function is not None:
        io_signals.append( signal_vector.find('h') )
    #
    io_signals[0].trajectory.extend([np.nan] *
                                    (len(dataset['current_values']) - no_of_initial_values))
    #
    signal_tuple = io_signals, bf_signal_vector.signals, hysteresis_bf_signal_vector.signals
    signal_trajectories = generate_signal_trajectories(signal_tuple, model.model_order,
                                                        no_of_initial_values)
    io_trajectories, p_trajectories, h_trajectories = signal_trajectories
    #
    simulated_voltage = Signal('v_sim', signal_vector.find('v').trajectory[:no_of_initial_values],
                               lambda x: x)
    combined_trajectories_dict = io_trajectories | p_trajectories | h_trajectories
    #
    for delay in range(1, model.model_order + 1):
        key = simulated_voltage.symbol + ('(k-' + str(delay) + ')' if delay > 0 else '(k)')
        #
        combined_trajectories_dict[key] = [simulated_voltage.trajectory[-delay]]

    for k in range(len(signal_vector.find('i').trajectory) - no_of_initial_values):
        #
        phi = []
        for key in model.model_terms:
            # Replace v with v_sim in the symbols
            new_symbols = []
            for sym in key.split('×'):
                sym_parts = sym.split('(')
                if sym_parts[0] == 'v':
                    new_symbols.append(simulated_voltage.symbol + '(' + sym_parts[1])
                else:
                    new_symbols.append(sym)
            # The value corresponding to each column for kth time instant.
            term_value = np.prod([combined_trajectories_dict[sym][k] for sym in new_symbols])
            phi.append(term_value)

        # Newly calculated voltage value at kth time instant.
        v_sim_new_value = np.dot( phi, model.model_estimate )
        # Make updates.
        simulated_voltage.update_trajectory(v_sim_new_value)
        for delay in range(1, model.model_order + 1):
            key = simulated_voltage.symbol + ('(k-' + str(delay) + ')' if delay > 0 else '(k)')
            #
            combined_trajectories_dict[key].append( simulated_voltage.trajectory[-delay] )

    emf_trajectory = []
    for i, s in enumerate(signal_vector.find('s').trajectory):
        t = signal_vector.find('T').trajectory[i] if 'temperature_values' in dataset else None
        emf_trajectory.append(model.emf_function(s, t))
    #
    return np.add(emf_trajectory, simulated_voltage.trajectory)



"""
Testing related to generation of basis functions from
user-provided strings.
"""

from pybatteryid.basisfunctions import Signal, generate_io_trajectories, \
    generate_basis_trajectories

def test_io_trajectories():
    """Test IO trajectories."""
    voltage = Signal('v', [1, 2, 3, 4, 5], lambda x: x)
    current = Signal('i', [10, 11, 12, 13, 14], lambda x: x)

    time_delays = [ [1, 2], [0, 1, 2] ]

    trajectories_dict = generate_io_trajectories([voltage, current], time_delays, 2)

    assert trajectories_dict['i(k)'] == [12, 13, 14]
    assert trajectories_dict['v(k-2)'] == [1, 2, 3]
    assert trajectories_dict['i(k-1)'] == [11, 12, 13]

def test_basis_trajectories():
    """Test basis trajectories."""
    soc = Signal('s', [1, 2, 3, 4, 5], lambda x: x)

    time_delays = [ [1, 2], [0, 1, 2] ]

    trajectories_dict = generate_basis_trajectories([soc], time_delays, 2)

    assert trajectories_dict['s(k)'] == [3, 4, 5]
    assert trajectories_dict['s(k-1)'] == [2, 3, 4]
    assert trajectories_dict['s(k-2)'] == [1, 2, 3]



"""
Contains utilities to generate current and temperature
profiles for model identification.
"""


import numpy as np


def generate_rectangular_pulse_train(pulse_periods):
    """Generate a rectangular pulse train."""
    # Time instants (with sampling period = 1 sec)
    time = np.arange(np.sum(pulse_periods))
    # Starting time instants of pulses
    pulse_start_times = np.cumsum(pulse_periods[:-1])
    pulse_start_times = np.insert(pulse_start_times, 0, 0)

    # Generate rectangular shape by setting 1s' between
    # starting times and half of the corresponding period.
    # Later, it is shifted downward by 0.5, and then multiplied
    # with 2 to scale the amplitudes to unity. The minus ensures
    # the discharge pulse phase at the beginning.
    rectangular_shape = (time >= pulse_start_times[:, np.newaxis]) & \
        (time < pulse_start_times[:, np.newaxis] + 0.5 * pulse_periods[:, np.newaxis])
    pulse_train = -2 * (np.sum(rectangular_shape, axis=0 ) - 0.5)

    return pulse_train


# pylint: disable=too-many-locals, too-many-arguments, R0917
def generate_current_profile(mean_pulse_period: float,
                             std_pulse_period: float,
                             mean_rest_period: float,
                             std_rest_period: float,
                             initial_amplitude: float,
                             amplitude_plus: float,
                             amplitude_minus: float,
                             std_random_amplitude: float,
                             dc_offset: float,
                             no_of_pulses: int,
                             sampling_frequency: float = 1,
                             upper_safety_limit: float | None = None,
                             lower_safety_limit: float | None = None):
    """Generate current profile used as identification input.
    
    Parameters:
    ----------
        mean_pulse_period : float
            `μ_tau` - Mean pulse period.
        std_pulse_period : float
            `σ_tau` - Standard deviation in the pulse period.
        mean_rest_period : float
            `μ_tau0` - Mean rest period.
        std_rest_period : float
            `σ_tau0` - Standard deviation in the rest period.
        initial_amplitude : float
            `alpha0` - Initial amplitude of the pulse train.
        amplitude_plus : float
            `alpha+` - Upper limit on the final amplitude `alpha`.
        amplitude_minus : float
            `alpha-` - Lower limit on the final amplitude `alpha`.
        std_random_amplitude : float
            `σ_xi` - Standard deviation in the randomness added to
            the signal.
        dc_offset : float
            `nu` - DC offset added to the pulse train.
        no_of_pulses : int
            Number of pulses in the signal.
        sampling_frequency : float
            Sampling frequency of the input signal.
        upper_safety_limit : float | None
            Upper limit on the identification signal for safety reasons.
        lower_safety_limit : float | None
            Lower limit on the identification signal for safety reasons.
    """
    # Shorthand
    fs = sampling_frequency
    # First, we generate periods for each pulse.
    pulse_periods = np.round(np.abs(np.random.default_rng().normal(mean_pulse_period * fs,
                                                                   std_pulse_period * fs,
                                                                   no_of_pulses)))
    # Generate pulse train.
    pulse_train = generate_rectangular_pulse_train(pulse_periods)

    # Split pulse train into list of phases
    phase_start_times = np.ceil(np.cumsum(np.repeat(pulse_periods, 2) / 2)).astype(int)
    phases = np.split(pulse_train, phase_start_times[:-1])

    # Add rest period following each phase
    rest_periods = np.rint(np.abs(np.random.default_rng().normal(mean_rest_period * fs,
                                                                 std_rest_period * fs,
                                                                 len(phases)))).astype(int)
    # Define the final amplitude of the signal
    # NOTE: This generates random number per pulse, and not per phase,
    # that is, we generate unique random numbers half the number of
    # phases, and then repeat them.
    final_amplitudes = np.repeat(np.random.default_rng().uniform(amplitude_minus,
                                                                 amplitude_plus,
                                                                 int(len(phases) / 2)), 2)
    # We finally construct the identification signal
    identification_signal = []
    for i, phase in enumerate(phases):
        # First, we define a sequence of zeros to add rest period
        # following each pulse.
        rest_signal = np.zeros(rest_periods[i])
        # Now, we add zero-mean random signal for high-frequency
        # excitations.
        zero_mean_random_signal = np.random.default_rng().normal(0,
                                                                 std_random_amplitude,
                                                                 len(phase))
        finalized_phase = final_amplitudes[i] * (
            initial_amplitude * phase + dc_offset + zero_mean_random_signal
        )
        # Applying safety limits if provided
        if upper_safety_limit is not None:
            finalized_phase[finalized_phase > upper_safety_limit] = upper_safety_limit
        if lower_safety_limit is not None:
            finalized_phase[finalized_phase < lower_safety_limit] = lower_safety_limit

        identification_signal = np.concatenate((identification_signal,
                                                finalized_phase,
                                                rest_signal))
    return identification_signal


def generate_temperature_profile(positive_phase_length,
                                 negative_phase_length,
                                 pulse_shift,
                                 pulse_amplitude,
                                 dc_offset,
                                 no_of_pulses):
    """Generate temperature profile used for model identification."""
    # First, we generate periods for each pulse.
    pulse_periods = [positive_phase_length, negative_phase_length] * int(no_of_pulses)
    # Generate pulse train.
    pulse_train = generate_rectangular_pulse_train(np.array(pulse_periods) * 2)

    # Split pulse train into list of phases
    phase_start_times = np.ceil(np.cumsum(np.repeat(pulse_periods, 2))).astype(int)
    phases = np.split(pulse_train, phase_start_times[:-1])[::2]

    # We finally construct the identification signal
    temperature_profile = []
    for i, phase in enumerate(phases):
        finalized_phase = phase
        finalized_phase *= -pulse_amplitude if i % 2 == 0 else pulse_amplitude
        finalized_phase += dc_offset

        temperature_profile = np.concatenate((temperature_profile, finalized_phase))

    # Remove the samples from the beginning of the
    # pulse train depending on `pulse_shift`
    temperature_profile = temperature_profile[int(pulse_shift * (positive_phase_length +
                                                                 negative_phase_length)):]

    return temperature_profile



"""
Testing related to extraction of basis functions from
user-provided strings.
"""

from pybatteryid.basisfunctions import Operation, extract_basis_functions

def test_no_operation_basis_function_extraction():
    """Test if the basis functions with no operation
    are extracted correctly"""

    strings = ['s']
    basis_functions = extract_basis_functions(strings)

    assert basis_functions[0].variable == 's'
    assert basis_functions[0].operation == Operation.NONE
    assert len(basis_functions[0].arguments) == 0
    assert basis_functions[0].function_string == 's'

def test_inverse_operation_basis_function_extraction():
    """Test if the basis functions with inverse operation
    are extracted correctly"""

    strings = ['1/s']
    basis_functions = extract_basis_functions(strings)

    assert basis_functions[0].variable == 's'
    assert basis_functions[0].operation == Operation.INVERSE
    assert len(basis_functions[0].arguments) == 0
    assert basis_functions[0].function_string == '1/s'

def test_logarithm_operation_basis_function_extraction():
    """Test if the basis functions with logarithm operation
    are extracted correctly"""

    strings = ['log[T]']
    basis_functions = extract_basis_functions(strings)

    assert basis_functions[0].variable == 'T'
    assert basis_functions[0].operation == Operation.LOGARITHM
    assert len(basis_functions[0].arguments) == 0
    assert basis_functions[0].function_string == 'log[T]'

def test_exp_sqrt_abs_operation_basis_function_extraction():
    """Test if the basis functions with exponential/square root/absolute
    operations are extracted correctly"""

    strings = ['exp[0.05*sqrt[|i|]]', 'exp[-2.759*sqrt[0.5*s-140]]']
    basis_functions = extract_basis_functions(strings)

    assert basis_functions[0].variable == 'i'
    assert basis_functions[0].operation == Operation.EXP_SQRT_ABS
    assert basis_functions[0].arguments == ['0.05', '', '', '|', '|', 'sqrt']
    assert basis_functions[0].function_string == 'exp[0.05*sqrt[|i|]]'

    assert basis_functions[1].variable == 's'
    assert basis_functions[1].operation == Operation.EXP_SQRT_ABS
    assert basis_functions[1].arguments == ['-2.759', '0.5', '-140', '', '', 'sqrt']
    assert basis_functions[1].function_string == 'exp[-2.759*sqrt[0.5*s-140]]'

def test_exp_power_abs_operation_basis_function_extraction():
    """Test if the basis functions with exponential/power/absolute
    operations are extracted correctly"""

    strings = ['exp[[0.00366*T+1]^-1]', 'exp[-2.759*[|-1.5*i+1890|]^+2.89]']
    basis_functions = extract_basis_functions(strings)

    assert basis_functions[0].variable == 'T'
    assert basis_functions[0].operation == Operation.EXP_POWER_ABS
    assert basis_functions[0].arguments == ['', '0.00366', '+1', '', '', '-1']
    assert basis_functions[0].function_string == 'exp[[0.00366*T+1]^-1]'

    assert basis_functions[1].variable == 'i'
    assert basis_functions[1].operation == Operation.EXP_POWER_ABS
    assert basis_functions[1].arguments == ['-2.759', '-1.5', '+1890', '|', '|', '+2.89']
    assert basis_functions[1].function_string == 'exp[-2.759*[|-1.5*i+1890|]^+2.89]'
